# -*- coding: utf-8 -*-
import falcon
from graceful.parameters import IntParam
from graceful.resources.base import BaseResource


class BaseMixin():
    """Base mixin class"""
    def handle(self, handler, req, resp, **kwargs):
        params = self.require_params(req)

        meta, content = self.require_meta_and_content(
            handler, params, **kwargs
        )
        self.make_body(resp, params, meta, content)
        return content


class RetrieveMixin(BaseMixin):
    """
    Add retrieve capabilities to resource with GET requests.
    """

    def retrieve(self, params, meta, **kwargs):
        """
        Retrieve object method handler

        Value returned by this handler will be included in response
        'content' section.

        Args:
            params (dict): dictionary of parsed parameters accordingly
               to definitions provided as resource class atributes.
            meta (dict): dictionary of meta parameters anything added
               to this dict will will be later included in response
               'meta' section. This can already prepopulated by method
               that calls this handler.
            kwargs (dict): dictionary of values retrieved from route url
               template by falcon. This is suggested way for providing
               resource identifiers.


        Returns:
            value to be included in response 'content' section

        """
        raise NotImplementedError("retrieve method not implemented")

    def on_get(self, req, resp, handler=None, **kwargs):
        self.handle(
            handler or self.retrieve, req, resp, **kwargs
        )


class ListMixin(BaseMixin):
    """
    Add list capabilities to resource with GET requests.
    """

    def list(self, params, meta, **kwargs):
        """
        List objects method handler

        Value returned by this handler will be included in response
        'content' section.

        Args:
            params (dict): dictionary of parsed parameters accordingly
               to definitions provided as resource class atributes.
            meta (dict): dictionary of meta parameters anything added
               to this dict will will be later included in response
               'meta' section. This can already prepopulated by method
               that calls this handler.
            kwargs (dict): dictionary of values retrieved from route url
               template by falcon. This is suggested way for providing
               resource identifiers.


        Returns:
            value to be included in response 'content' section

        """
        raise NotImplementedError("list method not implemented")

    def on_get(self, req, resp, handler=None, **kwargs):
        self.handle(
            handler or self.list, req, resp, **kwargs
        )


class DeleteMixin(BaseMixin):
    """
    Add delete capabilities to resource with DELETE requests.
    """

    def delete(self, params, meta, **kwargs):
        """
        Delete object method handler

        Args:
            params (dict): dictionary of parsed parameters accordingly
               to definitions provided as resource class atributes.
            meta (dict): dictionary of meta parameters anything added
               to this dict will will be later included in response
               'meta' section. This can already prepopulated by method
               that calls this handler.
            kwargs (dict): dictionary of values retrieved from route url
               template by falcon. This is suggested way for providing
               resource identifiers.

        Returns:
            value to be included in response 'content' section

        """
        raise NotImplementedError("delete method not implemented")

    def on_delete(self, req, resp, handler=None, **kwargs):
        self.handle(
            handler or self.delete, req, resp, **kwargs
        )

        resp.status = falcon.HTTP_ACCEPTED


class UpdateMixin(BaseMixin):
    """
    Add update capabilities to resource with PUT requests.
    """

    def update(self, params, meta, **kwargs):
        """
        Update object method handler

        Value returned by this handler will be included in response
        'content' section.

        Args:
            params (dict): dictionary of parsed parameters accordingly
               to definitions provided as resource class atributes.
            meta (dict): dictionary of meta parameters anything added
               to this dict will will be later included in response
               'meta' section. This can already prepopulated by method
               that calls this handler.
            kwargs (dict): dictionary of values retrieved from route url
               template by falcon. This is suggested way for providing
               resource identifiers.

        Returns:
            value to be included in response 'content' section

        """
        raise NotImplementedError("update method not implemented")

    def on_put(self, req, resp, handler=None, **kwargs):
        self.handle(
            handler or self.update, req, resp, **kwargs
        )
        resp.status = falcon.HTTP_ACCEPTED


class CreateMixin(BaseMixin):
    """
    Add create capabilities to resource with POST requests.
    """
    def create(self, params, meta, **kwargs):
        """
        Create object method handler

        Value returned by this handler will be included in response
        'content' section.

        Args:
            params (dict): dictionary of parsed parameters accordingly
               to definitions provided as resource class atributes.
            meta (dict): dictionary of meta parameters anything added
               to this dict will will be later included in response
               'meta' section. This can already prepopulated by method
               that calls this handler.
            kwargs (dict): dictionary of values retrieved from route url
               template by falcon. This is suggested way for providing
               resource identifiers.


        Returns:
            value to be included in response 'content' section

        """
        raise NotImplementedError("create method not implemented")

    def get_object_location(self, obj):
        """
        Return location path that will be included as Location header.

        This handler is optional.

        """
        raise NotImplementedError("update method not implemented")

    def on_post(self, req, resp, handler=None, **kwargs):
        obj = self.handle(
            handler or self.create, req, resp, **kwargs
        )
        try:
            resp.location = self.get_object_location(obj)
        except NotImplementedError:
            pass

        resp.status = falcon.HTTP_CREATED


class PaginatedMixin(BaseResource):
    """
    Add simple pagination capabilities to resource.

    This class provides two additional parameters with some default
    descriptions and ``add_pagination_meta`` method that can update
    meta with more useful pagination information.

    Example usage:

    .. code-block:: python

        from graceful.resources.mixins import PaginatedMixin
        from graceful.resources.generic import ListResource

        class SomeResource(PaginatedMixin, ListResource):

            def list(self, params, meta):
                # params has now 'page' and 'page_size' params that
                # can be used for offset&limit-like operations
                self.add_pagination_meta(params, meta)

                # ...

    """
    page_size = IntParam(
        details="""Specifies number of result entries in single response""",
        default='10'
    )
    page = IntParam(
        details="""Specifies number of results page for response.
        Page count starts from 0""",
        default='0',
    )

    def add_pagination_meta(self, params, meta):
        meta['page_size'] = params['page_size']
        meta['page'] = params['page']

        meta['prev'] = "page={0}&page_size={1}".format(
            params['page'] - 1, params['page_size']
        ) if meta['page'] > 0 else None

        meta['next'] = "page={0}&page_size={1}".format(
            params['page'] + 1, params['page_size']
        ) if meta.get('has_more', True) else None



# -*- coding: utf-8 -*-
from graceful.validators import min_validator, max_validator


class BaseField(object):
    """
    Base field class for subclassing. To create new field type
    subclass `BaseField` and implement following methods:

    - ``from_representation()``: converts representation (used in
      request/response body) to internal value.

    - ``to_representation()``: converts internal value to representation
      that will be used in response body.


    Args:

        details (str): human readable description of field (it will be used
           for describing resource on OPTIONS requests).

        label (str): human readable label of a field (it will be used for
           describing resource on OPTIONS requests).

           *Note: it is recommended to use field names that are
           self-explanatory intead of relying on field labels.*

        source (str): name of internal object key/attribute that will be
           passed to field on ``.to_representation()`` call. Special ``'*'``
           value is allowed that will pass whole object to field when making
           representation. If not set then default source will
           be a field name used as a serializer's attribute.

        validators (list): list of validator callables.

        many (bool): set to True if field is in fact a list of given type
          objects

        read_only (bool): True if field is read only and cannot be set/modified
           by POST and PUT requests

    Example:

    .. code-block:: python

        class BoolField(BaseField):
            def from_representation(self, data):
                if data in {'true', 'True', 'yes', '1', 'Y'}:
                    return True:
                elif data in {'false', 'False', 'no', '0', 'N'}:
                    return False:
                else:
                    raise ValueError(
                        "{data} is not valid boolean field".format(
                            data=data
                        )
                    )

            def to_representation(self, value):
                return ["True", "False"][value]

    """
    spec = None
    type = None

    def __init__(
            self,
            details,
            label=None,
            source=None,
            validators=None,
            many=False,
            read_only=False,
    ):

        self.label = label
        self.source = source
        self.details = details
        self.validators = validators or []
        self.many = many
        self.read_only = read_only

    def from_representation(self, data):
        raise NotImplementedError(
            "{cls}.from_representation() method not implemented".format(
                cls=self.__class__.__name__
            )
        )

    def to_representation(self, value):
        raise NotImplementedError(
            "{cls}.to_representation() method not implemented".format(
                cls=self.__class__.__name__
            )
        )

    def describe(self, **kwargs):
        """
        Describe this field instance for purpose of self-documentation.

        Args:
            kwargs (dict): dictionary of additional description items for
               extending default description

        Returns:
            dict: dictionary of description items

        Suggested way for overriding description fields or extending it with
        additional items is calling super class method with new/overriden
        fields passed as keyword arguments like following:

        .. code-block:: python

            class DummyField(BaseField):
               def description(self, **kwargs):
                   super(DummyParam, self).describe(is_dummy=True, **kwargs)

        """
        description = {
            'label': self.label,
            'details': self.details,
            'type': "list of {}".format(self.type) if self.many else self.type,
            'spec': self.spec,
            'read_only': self.read_only,
        }
        description.update(kwargs)
        return description

    def validate(self, value):
        """ Perform validation on value by running all field validators.
        Single validator is a callable that accepts one positional argument
        and raises "ValidationError" when validation fails.

        Error message included in exception will be included in http error
        response

        Args:
            value: internal value to validate

        Returns:
            None

        Note:
            Concept of validation for fields is understood here as a process
            of checking if data of valid type (successfully parsed/processed by
            ``.from_representation`` handler) does meet some other constraints
            (lenght, bounds, unique, etc). Becasue of that this method is
            always called with result of ``.from_representation()`` passed
            as value.

        """
        for validator in self.validators:
            validator(value)


class RawField(BaseField):
    """
    Represents raw field subtype. Any value from resource object
    will be returned as is without any conversion and no control
    over serialized value type is provided. Can be used only with
    very simple data types like int, float, str etc. but can eventually
    cause problems if value provided in representation has type
    that is not accepted in application.

    Effect of using this can differ between various content-types.

    """
    type = 'raw'

    def from_representation(self, data):
        return data

    def to_representation(self, value):
        return value


class StringField(BaseField):
    """
    Represents string field subtype without any extensive validation.
    """
    type = 'raw'

    def from_representation(self, data):
        return str(data)

    def to_representation(self, value):
        return str(value)


class BoolField(BaseField):
    """
    Represents boolean type of field. By default accepts a wide range of
    incoming True/False representations:

    * False: ``[False', 'false', 'FALSE', 'F', 'f' '0', 0, 0.0, False``
    * True: ``['True', 'true', 'TRUE', 'T', 't', '1', 1, True]``

    By default by as representations of internal object's value it returns
    python's False/True values that will be later serialized to form that
    is native for content-type of use.

    This behavior can be changed using ``representations`` field argument.

    Args:

        representations (tuple): two-tuple with representations for
           (False, True) values, that will be used instead of default values

    """
    type = 'bool'

    _TRUE_VALUES = {'True', 'true', 'TRUE', 'T', 't', '1', 1, True}
    _FALSE_VALUES = {'False', 'false', 'FALSE', 'F', 'f' '0', 0, 0.0, False}

    _DEFAULT_REPRESENTATIONS = (False, True)

    def __init__(
            self,
            details,
            representations=None,
            **kwargs
    ):
        super(BoolField, self).__init__(details, **kwargs)

        if representations:
            # could not resist...
            self._FALSE_VALUES = {representations[False]}
            self._TRUE_VALUES = {representations[True]}

        self.representations = representations or self._DEFAULT_REPRESENTATIONS

    def from_representation(self, data):
        if data in self._TRUE_VALUES:
            return True
        elif data in self._FALSE_VALUES:
            return False
        else:
            raise ValueError(
                "{type} type value must be one of {{values}}".format(
                    type=self.type,
                    values=self._TRUE_VALUES.union(self._FALSE_VALUES)
                )
            )

    def to_representation(self, value):
        return self.representations[value]


class IntField(BaseField):
    """
    Represents integer type of field. Accepts both integers and strings
    as an incoming integer representation and always returns int as a
    representation of internal objects's value that will be later
    serialized to form that is native for content-type of use.

    This field accepts optional arguments that simply add new `max` and `min`
    value validation.

    Args:
        max_value (int): optional max value for validation
        min_value (int): optional min value for validation

    """
    type = 'int'

    def __init__(
            self,
            details,
            max_value=None,
            min_value=None,
            **kwargs
    ):
        super(IntField, self).__init__(details, **kwargs)

        self.max_value = max_value
        self.min_value = min_value

        if max_value is not None:
            self.validators.append(max_validator(max_value))
        if min_value is not None:
            self.validators.append(min_validator(min_value))

    def to_representation(self, value):
        return int(value)

    def from_representation(self, data):
        return int(data)


class FloatField(BaseField):
    """
    Represents float type of field. Accepts both floats and strings
    as an incoming float number representation and always returns float as a
    representation of internal objects's value that will be later
    serialized to form that is native for content-type of use.

    This field accepts optional arguments that simply add new `max` and `min`
    value validation.

    Args:
        max_value (int): optional max value for validation
        min_value (int): optional min value for validation

    """
    type = 'float'

    def __init__(
            self,
            details,
            max_value=None,
            min_value=None,
            **kwargs
    ):
        super(FloatField, self).__init__(details, **kwargs)

        self.max_value = max_value
        self.min_value = min_value

        if max_value is not None:
            self.validators.append(max_validator(max_value))
        if min_value is not None:
            self.validators.append(min_validator(min_value))

    def to_representation(self, value):
        return float(value)

    def from_representation(self, data):
        return float(data)



# -*- coding: utf-8 -*-
import re

from graceful.errors import ValidationError


__all__ = [
    'min_validator',
    'max_validator',
    'choices_validator',
]


def min_validator(min_value):
    """
    Return validator function that will check internal value with
    ``value >= min_value`` check

    Args:
        min_value: minimal value for new validator

    """
    def validator(value):
        if value < min_value:
            raise ValidationError("{} is not >= {}".format(value, min_value))

    return validator


def max_validator(max_value):
    """
    Return validator function that will check if ``value >= min_value``.

    Args:
        max_value: maximum value for new validator

    """
    def validator(value):
        if value > max_value:
            raise ValidationError("{} is not <= {}".format(value, max_value))

    return validator


def choices_validator(choices):
    """
    Return validator function that will check if ``value in choices``.

    Args:
        max_value (list, set, tuple): allowed choices for new validator

    """
    def validator(value):
        if value not in choices:
            # note: make it a list for consistent representation
            raise ValidationError(
                "{} is not in {}".format(value, list(choices))
            )

    return validator


def match_validator(expression):
    """
    Return validator function that will check if matches given match.

    Args:
        match: if string then this will be converted to regular expression
           using ``re.compile``. Can be also any object that has ``match()`
           method like already compiled regular regular expression or custom
           matching object/class.

    """

    if isinstance(expression, str):
        compiled = re.compile(expression)
    elif hasattr(expression, 'match'):
        # check it early so we could say something is wrong early
        compiled = expression
    else:
        raise TypeError(
            'Provided match is nor a string nor has a match method '
            '(like re expressions)'
        )

    def validator(value):
        if not compiled.match(value):
            # note: make it a list for consistent representation
            raise ValidationError(
                "{} does not match pattern: {}".format(
                    value,
                    compiled.pattern
                    if hasattr(compiled, 'pattern')
                    else compiled
                )
            )

    return validator



# -*- coding: utf-8 -*-
from functools import partial

from graceful.resources.base import BaseResource
from graceful.resources.mixins import (
    RetrieveMixin,
    ListMixin,
    UpdateMixin,
    CreateMixin,
    DeleteMixin,
    PaginatedMixin,
)


class Resource(RetrieveMixin, BaseResource):
    """
    Generic resource for single object retrieve type of endpoints without use
    of automatic representation serialization and extensive field descriptions.

    This still gives support for defining parameters.

    Example usage:

    .. code-block:

        from graceful.resources.generic import Resource
        from graceful.parameters import StringParam

        class SampleResource(Resource):
            dummy = StringParam("some example dummy parameter")

            def retrieve(self, params, meta):
                return {"sample": "resource"}

    """


class ListResource(ListMixin, BaseResource):
    """
    Generic resource for list object type of endpoints without use of
    automatic representation serialization and extensive field descriptions

    This still gives support for defining parameters.

    Example usage:

    .. code-block:

        from graceful.resources.generic import ListResource
        from graceful.parameters import StringParam

        class SampleResource(ListResource):
            some_filter = StringParam("some example filter parameter")

            def list(self, params, meta):
                return [{"sample": "resource"}]

    """
    pass


class RetrieveAPI(RetrieveMixin, BaseResource):
    """
    Generic resource that uses serializer for resource description,
    serialization and validation.

    Allowed methods:

    * GET: retrieve resource representation (handled with ``.retrieve()``
      method handler)

    """
    serializer = None

    def describe(self, req, resp, **kwargs):
        return super(RetrieveAPI, self).describe(
            req, resp,
            type='object',
            fields=self.serializer.describe() if self.serializer else None,
            **kwargs
        )

    def _retrieve(self, params, meta, **kwargs):
        return self.serializer.to_representation(
            self.retrieve(params, meta, **kwargs)
        )

    def on_get(self, req, resp, **kwargs):
        return super(RetrieveAPI, self).on_get(
            req, resp, handler=self._retrieve, **kwargs
        )


class RetrieveUpdateAPI(UpdateMixin, RetrieveAPI):
    """
    Generic resource that uses serializer for resource description,
    serialization and validation.

    Allowed methods:

    * GET: retrieve resource representation handled with ``.retrieve()``
      method handler
    * PUT: update resource with representation provided in request body
      (handled with ``.update()`` method handler)

    """
    def _update(self, params, meta, **kwargs):
        return self.serializer.to_representation(
            self.update(params, meta, **kwargs)
        )

    def on_put(self, req, resp, **kwargs):
        validated = self.require_validated(req)
        return super(RetrieveUpdateAPI, self).on_put(
            req, resp,
            handler=partial(self._update, validated=validated),
            **kwargs
        )


class RetrieveUpdateDeleteAPI(DeleteMixin, RetrieveUpdateAPI):
    """
    Generic resource that uses serializer for resource description,
    serialization and validation.

    Allowed methods:

    * GET: retrieve resource representation (handled with ``.retrieve()``
      method handler)
    * PUT: update resource with representation provided in request body
      (handled with ``.update()`` method handler)
    * DELETE: delete resource (handled with ``.delete()`` method handler)

    """


class ListAPI(ListMixin, BaseResource):
    """
    Generic resource that uses serializer for resource description,
    serialization and validation.

    Allowed methods:

    * GET: list multiple resource instances representations (handled
      with ``.list()`` method handler)

    """
    def _list(self, params, meta, **kwargs):
        return [
            self.serializer.to_representation(obj)
            for obj in self.list(params, meta, **kwargs)
        ]

    def describe(self, req, resp, **kwargs):
        return super(ListAPI, self).describe(
            req, resp,
            type='list',
            fields=self.serializer.describe() if self.serializer else None,
            **kwargs
        )

    def on_get(self, req, resp, **kwargs):
        return super(ListAPI, self).on_get(req, resp, handler=self._list)


class ListCreateAPI(CreateMixin, ListAPI):
    """
    Generic resource that uses serializer for resource description,
    serialization and validation.

    Allowed methods:

    * GET: list multiple resource instances representations (handled
      with ``.list()`` method handler)
    * POST: create new resource from representation provided in request body
      (handled with ``.create()`` method handler)

    """
    def _create(self, params, meta, **kwargs):
        return self.serializer.to_representation(
            self.create(params, meta, **kwargs)
        )

    def on_post(self, req, resp, **kwargs):
        validated = self.require_validated(req)

        return super(ListCreateAPI, self).on_post(
            req, resp,
            handler=partial(self._create, validated=validated),
            **kwargs
        )


class PaginatedListAPI(PaginatedMixin, ListAPI):
    """
    Generic resource that uses serializer for resource description,
    serialization and validation.

    Adds simple pagination to list of resources.

    Allowed methods:

    * GET: list multiple resource instances representations (handled
      with ``.list()`` method handler)

    """
    def _list(self, params, meta, **kwargs):
        objects = super(PaginatedListAPI, self)._list(params, meta, **kwargs)
        # note: we need to populate meta after objects are retrieved
        self.add_pagination_meta(params, meta)
        return objects


class PaginatedListCreateAPI(PaginatedMixin, ListCreateAPI):
    """
    Generic resource that uses serializer for resource description,
    serialization and validation.

    Adds simple pagination to list of resources.

    Allowed methods:

    * GET: list multiple resource instances representations (handled
      with ``.list()`` method handler)
    * POST: create new resource from representation provided in request body
      (handled with ``.create()`` method handler)

    """
    def _list(self, params, meta, **kwargs):
        objects = super(PaginatedListCreateAPI, self)._list(
            params, meta, **kwargs
        )
        # note: we need to populate meta after objects are retrieved
        self.add_pagination_meta(params, meta)
        return objects



# -*- coding: utf-8 -*-
import base64
import decimal
import binascii
import inspect


class BaseParam():
    """
    Base parameter class for subclassing. To create new parameter type
    subclass ``BaseField`` and implement following methods:

    To create new field type subclass ``BaseParam`` and implement ``.value()``
    method handlers.


    Args:

        details (str): verbose description of parameter. Should contain all
           information that may be important to your API user and will be used
           for describing resource on ``OPTIONS`` requests and ``.describe()``
           call.

        label (str): human readable label for this parameter (it will be used
           for describing resource on OPTIONS requests).

           *Note that it is recomended to use parameter names that are
           self-explanatory intead of relying on param labels.*

        required (bool): if set to ``True`` then all GET, POST, PUT,
           PATCH and DELETE requests will return ``400 Bad Request`` response
           if query param is not provided.

        default (str): set default value for param if it is not
           provided in request as query parameter. This MUST be a raw string
           value that will be then parsed by ``.value()`` handler.

           If default is set and ``required`` is ``True`` it will raise
           ``ValueError`` as having required parameters with default
           value has no sense.

    Example:

    .. code-block:: python

           class BoolParam(BaseParam):
               def value(self, data):
                   if data in {'true', 'True', 'yes', '1', 'Y'}:
                       return True:
                   elif data in {'false', 'False', 'no', '0', 'N'}:
                       return False:
                   else:
                       raise ValueError(
                           "{data} is not valid boolean field".format(
                               data=data
                           )
                       )

    """
    spec = None
    type = None

    def __init__(
            self,
            details,
            label=None,
            required=False,
            default=None,
    ):
        self.label = label
        self.details = details
        self.required = required

        if not (default is None) and not isinstance(default, str):
            raise TypeError(
                "value for {cls} 'default' argument must be string instance"
                "".format(cls=self.__class__.__name__)
            )

        if not (default is None) and required:
            raise ValueError(
                "{cls}(required={required}, default='{default}'): "
                "initialization with both required and default makes no sense"
                "".format(
                    cls=self.__class__.__name__,
                    default=default,
                    required=required,
                )
            )
        self.default = default

    def value(self, raw_value):
        """
        Raw value deserializtion method handler

        Args:
            raw_value (str) - raw value from GET parameters

        """
        raise NotImplementedError(
            "{cls}.value() method not implemented".format(
                cls=self.__class__.__name__
            )
        )

    def describe(self, **kwargs):
        """
        Describe this parameter instance for purpose of self-documentation.

        Args:
            kwargs (dict): dictionary of additional description items for
               extending default description

        Returns:
            dict: dictionary of description items


        Suggested way for overriding description fields or extending it with
        additional items is calling super class method with new/overriden
        fields passed as keyword arguments like following:

        .. code-block:: python

            class DummyParam(BaseParam):
               def description(self, **kwargs):
                   super(DummyParam, self).describe(is_dummy=True, **kwargs)

        """

        description = {
            'label': self.label,
            # note: details are expected to be large so it should
            #       be reformatted
            'details': inspect.cleandoc(self.details),
            'required': self.required,
            'spec': self.spec,
            'default': self.default,
            'type': self.type or 'unspecified'
        }

        description.update(kwargs)
        return description


class StringParam(BaseParam):
    """
    Describes parameter that will always be returned in same form as provided
    in query string. Still additional validation can be added to param instance
    e.g.:

    .. code-block:: python

        from graceful.parameters import StringParam
        from graceful.validators import match_validator
        from graceful.resources.generic import Resource

        class ExampleResource(Resource):
            word = StringParam(
                'one "word" parameter',
                validators=[match_validator('\w+')],
            )

    """
    type = 'string'

    def value(self, raw_value):
        """Returns value as-is (str)"""
        return raw_value


class Base64EncodedParam(BaseParam):
    """
    Describes string parameter that has value encoded using Base64 encoding
    """
    spec = (
        "RFC-4648 Section 4",
        "https://tools.ietf.org/html/rfc4648#section-4",
    )

    def value(self, raw_value):
        """Decodes param with Base64"""
        try:
            return base64.b64decode(bytes(raw_value, 'utf-8')).decode('utf-8')
        except binascii.Error as err:
            raise ValueError(str(err))


class IntParam(BaseParam):
    """
    Describes parameter that has value expressed as integer number
    """
    type = "integer"

    def value(self, raw_value):
        """Decodes param as integer value"""
        return int(raw_value)


class FloatParam(BaseParam):
    """
    Describes parameter that has value expressed as float number
    """
    type = "float"

    def value(self, raw_value):
        """Decodes param as float value"""
        return float(raw_value)


class DecimalParam(BaseParam):
    """
    Describes parameter that has value expressed as decimal number
    """
    type = "decimal"

    def value(self, raw_value):
        """Decodes param as decimal value"""
        try:
            return decimal.Decimal(raw_value)
        except decimal.InvalidOperation:
            raise ValueError(
                "Could not parse '{}' value as decimal".format(raw_value)
            )



# -*- coding: utf-8 -*-
from setuptools import setup, find_packages
import os


def get_version(version_tuple):
    if not isinstance(version_tuple[-1], int):
        return '.'.join(map(str, version_tuple[:-1])) + version_tuple[-1]
    return '.'.join(map(str, version_tuple))


init = os.path.join(
    os.path.dirname(__file__),
    'src', 'graceful', '__init__.py'
)
version_line = list(filter(lambda l: l.startswith('VERSION'), open(init)))[0]
VERSION = get_version(eval(version_line.split('=')[-1]))

INSTALL_REQUIRES = [
    'falcon'
]

try:
    from pypandoc import convert

    def read_md(f):
        return convert(f, 'rst')

except ImportError:
    convert = None
    print(
        "warning: pypandoc module not found, could not convert Markdown to RST"
    )

    def read_md(f):
        return open(f, 'r').read()  # noqa

README = os.path.join(os.path.dirname(__file__), 'README.md')
PACKAGES = find_packages('src')
PACKAGE_DIR = {'': 'src'}

setup(
    name='graceful',
    version=VERSION,
    author='Michał Jaworski',
    author_email='swistakm@gmail.com',
    description='falcon REST done with grace',
    long_description=read_md(README),
    packages=PACKAGES,
    package_dir=PACKAGE_DIR,

    url='https://github.com/swistakm/graceful',
    include_package_data=True,
    install_requires=INSTALL_REQUIRES,
    zip_safe=True,

    license="BSD",
    classifiers=[
        'Development Status :: 4 - Beta',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: BSD License',
        'Operating System :: OS Independent',
        'Topic :: Software Development :: Libraries :: Application Frameworks',

        'Programming Language :: Python',
        'Programming Language :: Python :: 3.2',
        'Programming Language :: Python :: 3.3',
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3 :: Only',
        'Topic :: Internet :: WWW/HTTP :: WSGI',
    ],
)



# -*- coding: utf-8 -*-
from collections import OrderedDict
from collections.abc import Mapping, MutableMapping

from graceful.errors import DeserializationError, ValidationError
from graceful.fields import BaseField


class MetaSerializer(type):
    """ Metaclass for handling serialization with field objects
    """
    _fields_storage_key = '_fields'

    @classmethod
    def __prepare__(mcs, name, bases, **kwargs):
        """
        Prepare class namespace in a way that ensures order of attributes.
        This needs to be an `OrderedDict` so `_get_params()` method can
        construct fields storage that preserves the same order of fields as
        defined in code.

        Note: this is python3 thing and support for ordering of params in
        descriptions will not be backported to python2 even if this framework
        will get python2 support.

        """
        return OrderedDict()

    @classmethod
    def _get_fields(mcs, bases, namespace):
        """ Pop all field objects from attributes dict (namespace)
        and store them under _field_storage_key atrribute.
        Also collect all fields from base classes in order that ensures
        fields can be overriden.

        Args:
            bases: all base classes of created serializer class
            namespace (dict): namespace as dictionary of attributes

        """
        fields = [
            (name, namespace.pop(name))
            for name, attribute
            in list(namespace.items())
            if isinstance(attribute, BaseField)
        ]

        for base in reversed(bases):
            if hasattr(base, mcs._fields_storage_key):
                fields = list(
                    getattr(base, mcs._fields_storage_key).items()
                ) + fields

        return OrderedDict(fields)

    def __new__(mcs, name, bases, namespace):
        namespace[mcs._fields_storage_key] = mcs._get_fields(bases, namespace)
        return super(MetaSerializer, mcs).__new__(
            # note: there is no need preserve order in namespace anymore so
            # we convert it explicitely to dict
            mcs, name, bases, dict(namespace)
        )


class BaseSerializer(object, metaclass=MetaSerializer):
    """
    Base serializer class for describing internal object serialization

    Example:

    .. code-block:: python

        from graceful.serializers import BaseSerializer
        from graceful.fields import RawField, IntField, FloatField


        class CatSerializer(BaseSerializer):
            species = RawField("non normalized cat species")
            age = IntField("cat age in years")
            height = FloatField("cat height in cm")

    """

    @property
    def fields(self):
        """
        Dictionary of field objects defined for this resource serialization
        """
        return getattr(self, self.__class__._fields_storage_key)

    def to_representation(self, obj):
        """
        Convert given internal object instance into defined representation
        that will be later serialized to content-type of use in resource
        http method handler.

        This loops over all fields and retrieves source keys/attributes as
        field values with respect to optional field sources and converts each
        one using ``field.to_representation()`` method.

        Args:
            obj (object): internal object that needs to be represented

        Returns:
            dict: representation dictionary

        """
        representation = {}

        for name, field in self.fields.items():
            # note fields do not know their names in source representation
            # but may know what attribute they target from source object
            attribute = self.get_attribute(obj, field.source or name)

            if attribute is None:
                # Skip none attributes so fields do not have to deal with them
                representation[name] = [] if field.many else None
            elif field.many:
                representation[name] = [
                    field.to_representation(item) for item in attribute
                ]
            else:
                representation[name] = field.to_representation(attribute)

        return representation

    def from_representation(self, representation):
        """
        Convert given representation dict into dictionary of internal object
        values with respect to field sources.

        This does not check if all required fields exist or values are
        valid in terms of value validation
        (see: :meth:`BaseField.validate()`) but still requires all of passed
        representation values to be well formed representation (success call
        to ``field.from_representation``).

        In case of malformed representation it will run additional validation
        only to provide a full detailed exception about all that might be
        wrong with provided representation.

        Args:
           representation (dict): dictionary with field representation values

        Raises:
            DeserializationError: when at least one representation field
               is not formed as expected by field object. Information
               about additional forbidden/missing/invalid fields is provided
               as well.

        """
        object_dict = {}
        failed = {}

        for name, field in self.fields.items():
            if name not in representation:
                continue

            try:
                # if field has explicitely specified source then use it
                # else fallback to field name.
                # Note: field does not know its name
                object_dict[field.source or name] = field.from_representation(
                    representation[name]
                )
            except ValueError as err:
                failed[name] = str(err)

        if failed:
            # if failed to parse we eagerly perform validation so full
            # information about what is wrong will be returned
            try:
                self.validate(object_dict)
                # note: this exception can be reached with partial==True
                # since do not support partial updates yet this has 'no cover'
                raise DeserializationError()  # pragma: no cover
            except DeserializationError as err:
                err.failed = failed
                raise

        return object_dict

    def validate(self, object_dict, partial=False):
        """
        Validate given internal object agains missing/forbidden/invalid
        fields values using fields definitions defined in serializer.

        Args:
            object_dict (dict): internal object dictionart to perform
              to validate
            partial (bool): if set to True then incomplete object_dict
              is accepter and will not raise any exceptions when one
              of fields is missing

        Raises:
            DeserializationError:

        """

        # we are working on object_dict not an representation so there
        # is a need to annotate sources differently
        sources = {
            # TODO: handling of '*' sources here is a bit terryfying
            # TODO: maybe this needs more care in future releases
            field.source or name if field.source != "*" else name: field
            for name, field in self.fields.items()
        }

        # note: we are checking for all mising and invalid fields so we can
        # return exception with all fields that are missing and should
        # exist instead of single one
        missing = [
            name for name, field in sources.items()
            if all((not partial, name not in object_dict, not field.read_only))
        ]

        forbidden = [
            name for name in object_dict
            if any((name not in sources, sources[name].read_only))
        ]

        invalid = {}
        for name, value in object_dict.items():
            try:
                sources[name].validate(value)
            except ValidationError as err:
                invalid[name] = str(err)

        if any([missing, forbidden, invalid]):
            raise DeserializationError(missing, forbidden, invalid)

    def get_attribute(self, obj, attr):
        """
        Get attribute from given object instance where 'attribute' can
        be also a key from object if is a dict or any kind of mapping

        Note: it will return None if attribute key does not exist

        Args:
            obj (object): internal object to retrieve data from

        Returns:
            internal object's key value or attribute

        """
        # '*' is a special wildcard character that means whole object
        # is passed
        if attr == '*':
            return obj

        # if this is any mapping then instead of attributes use keys
        if isinstance(obj, Mapping):
            return obj.get(attr, None)

        return getattr(obj, attr, None)

    def set_attribute(self, obj, attr, value):
        """
        Set attribute in given object instance where 'attribute' can
        be also a key from object if it is a dict or any kind of mapping

        Args:
            obj (object): object instance to modify
            attr (str): attribute (or key) to change
            value: value to set

        """
        # if this is any mutable mapping then instead of attributes use keys
        if isinstance(obj, MutableMapping):
            obj[attr] = value
        else:
            setattr(obj, attr, value)

    def describe(self):
        """
        Describe whole all fields defined for this serializer using their own
        descriptions with respect to order in which they are defined as class
        attributes.

        Returns:
            OrderedDict: serializer description

        """
        return OrderedDict([
            (name, field.describe())
            for name, field in self.fields.items()
        ])



# -*- coding: utf-8 -*-
from falcon import HTTPBadRequest


class DeserializationError(ValueError):
    """
    Raised when error happened during deserialization of object
    """
    def __init__(
            self, missing=None, forbidden=None, invalid=None, failed=None
    ):
        self.missing = missing
        self.forbidden = forbidden
        self.invalid = invalid
        self.failed = failed

    def as_bad_request(self):
        return HTTPBadRequest(
            title="Representation deserialization failed",
            description=self._get_description()
        )

    def _get_description(self):
        """
        Return human readable description that explains everything that
        went wrong with deserialization.

        """
        return ", ".join([
            part for part in [
                "missing: {}".format(self.missing) if self.missing else "",
                (
                    "forbidden: {}".format(self.forbidden)
                    if self.forbidden else ""
                ),
                "invalid: {}:".format(self.invalid) if self.invalid else "",
                (
                    "failed to parse: {}".format(self.failed)
                    if self.failed else ""
                )
            ] if part
        ])


class ValidationError(ValueError):
    """Raised when validation error occured"""



# -*- coding: utf-8 -*-
VERSION = (0, 0, 2)  # PEP 386  # noqa
__version__ = ".".join([str(x) for x in VERSION])  # noqa

"""
Minimalist framework for self-descriptive RESTful APIs build on top of
falcon.

It is inspired by Django REST Framework package. Mostly by how object
serialization is done but more emphasis is put on API to being
self-descriptive.

"""
